#lang racket

(require "CCA.rkt"
         (prefix-in prototype: "define-judgment-form/define-judgment-form.rkt")
         redex/reduction-semantics)

(provide typeof typeof+
         to-type-form
         from-type-form)

(define-syntax-rule (define-CCA-typing-extension typeof . extension)
  (prototype:define-judgment-form
    ; core forms
    ((typeof (? Γ) (unit) unit)
     "unit")
    ((bound (? x) (? Γ) (? τ))
     (typeof (? Γ) (var (? x)) (? τ))
     "var")
    ((typeof ((? τ_1) (? Γ)) (? e) (? τ_2))
     (typeof (? Γ) (λ (? e)) (-> (? τ_1) (? τ_2)))
     "abs")
    ((typeof (? Γ) (? e_1) (-> (? τ_2) (? τ)))
     (typeof (? Γ) (? e_2) (? τ_2))
     (typeof (? Γ) (app (? e_1) (? e_2)) (? τ))
     "app")
    ((typeof (? Γ) (? e_1) (? τ_1))
     (typeof (? Γ) (? e_2) (? τ_2))
     (typeof (? Γ) (pair (? e_1) (? e_2)) (pair (? τ_1) (? τ_2)))
     "pair")
    ((typeof (? Γ) (? e) (pair (? τ_1) (? τ_2)))
     (typeof (? Γ) (fst (? e)) (? τ_1))
     "fst")
    ((typeof (? Γ) (? e) (pair (? τ_1) (? τ_2)))
     (typeof (? Γ) (snd (? e)) (? τ_2))
     "snd")
    
    ; numbers
    ((typeof (? Γ) (num (? n)) num)
     "number")
    ((typeof (? Γ) + (-> num (-> num num)))
     "+")
    ((typeof (? Γ) * (-> num (-> num num)))
     "*")
    ((typeof (? Γ) / (-> num (-> num num)))
     "/")
    
    ; arrows
    ((typeof (? Γ) arr (-> (-> (? α) (? β))
                         (arr (? α) (? β))))
     "arr")
    ((typeof (? Γ) >>> (-> (arr (? α) (? β))
                         (-> (arr (? β) (? θ))
                             (arr (? α) (? θ)))))
     ">>>")
    ((typeof (? Γ) first (-> (arr (? α) (? β))
                           (arr (pair (? α) (? θ))
                                (pair (? β) (? θ)))))
     "first")
    ((typeof (? Γ) second (-> (arr (? α) (? β))
                            (arr (pair (? θ) (? α))
                                 (pair (? θ) (? β)))))
     "second")
    ((typeof (? Γ) loop (-> (arr (pair (? α) (? θ))
                               (pair (? β) (? θ)))
                          (arr (? α) (? β))))
     "loop")
    ((typeof (? Γ) init (-> (? α) (arr (? α) (? α))))
     "init")
    . extension))

(define-CCA-typing-extension typeof)

(prototype:define-judgment-form
  ((bound z ((? τ) (? Γ)) (? τ))
   "bound-eq")
  ((bound (? x) (? Γ) (? τ))
   (bound (s (? x)) ((? σ) (? Γ)) (? τ))
   "bound-neq"))

(define (to-type-form expr [env '()])
  (define (to-peano n)
    (for/fold ([p 'z]) ([_ (in-range n)])
      `(s ,p)))
  (define-metafunction CCA
    [(translate x_i (x_0 ... x_i x_i+1 ...))
     (var ,(to-peano (length (term (x_0 ...)))))]
    [(translate (λ (x) e) (x_0 ...))
     (λ (translate e (x x_0 ...)))]
    [(translate (e_1 e_2) (x_0 ...))
     (app (translate e_1 (x_0 ...))
          (translate e_2 (x_0 ...)))]
    [(translate () (x_0 ...))
     (unit)]
    [(translate number (x_0 ...))
     (num number)]
    [(translate (any ...) (x_0 ...))
     ((translate any (x_0 ...)) ...)]
    [(translate any (x_0 ...))
     any])
  (term (translate ,expr ,env)))

(define (from-type-form expr)
  (define from-peano
    (match-lambda
      ['z 0]
      [`(s ,n) (add1 (from-peano n))]))
  (let translate ([e expr] [d 0])
    (match e
      [`(var ,i)
       (string->symbol (format "x~s" (- d (from-peano i))))]
      [`(λ ,e)
       `(λ (,(string->symbol (format "x~s" (add1 d))))
          ,(translate e (add1 d)))]
      [`(app ,e1 ,e2)
       `(,(translate e1 d) ,(translate e2 d))]
      [`(unit) `()]
      [`(num ,n) n]
      [(? list?)
       (for/list ([x e])
         (translate x d))]
      [x x])))

(define-CCA-typing-extension typeof+
  [(typeof+ (? Γ) assoc
            (-> (pair (pair (? α) (? β)) (? θ))
                (pair (? α) (pair (? β) (? θ)))))
   "assoc"]
  [(typeof+ (? Γ) assoc-inv
            (-> (pair (? α) (pair (? β) (? θ)))
                (pair (pair (? α) (? β)) (? θ))))
   "assoc-inv"]
  [(typeof+ (? Γ) swap (-> (pair (? α) (? β)) (pair (? β) (? α))))
   "swap"]
  [(typeof+ (? Γ) times
            (-> (-> (? α) (? β))
                (-> (-> (? θ) (? γ))
                    (-> (pair (? α) (? θ))
                        (pair (? β) (? γ))))))
   "times"]
  [(typeof+ (? Γ) compose
            (-> (-> (? β) (? θ))
                (-> (-> (? α) (? β))
                    (-> (? α) (? θ)))))
   "compose"]
  [(typeof+ (? Γ) id (-> (? α) (? α)))
   "id"]
  [(typeof+ (? Γ) juggle
            (-> (pair (pair (? α) (? β)) (? θ))
                (pair (pair (? α) (? θ)) (? β))))
   "juggle"]
  [(typeof+ (? Γ) shuffle
            (-> (pair (? α) (pair (pair (? β) (? δ)) (pair (? θ) (? η))))
                (pair (pair (? α) (pair (? β) (? θ))) (pair (? δ) (? η)))))
   "shuffle"]
  [(typeof+ (? Γ) shuffle-inv
            (-> (pair (pair (? α) (pair (? β) (? θ))) (pair (? δ) (? η)))
                (pair (? α) (pair (pair (? β) (? δ)) (pair (? θ) (? η))))))
   "shuffle-inv"]
  [(typeof+ (? Γ) second
            (-> (arr (? α) (? β))
                (arr (pair (? θ) (? α))
                     (pair (? θ) (? β)))))
   "second"]
  [(typeof+ (? Γ) loopB
            (-> (? θ)
                (-> (arr (pair (? α) (pair (? γ) (? θ)))
                         (pair (? β) (pair (? γ) (? θ))))
                    (arr (? α) (? β)))))
   "loopB"])