#lang scheme

(require redex/reduction-semantics
         "test-common.rkt"
         "typing.rkt"
         "reduction.rkt"
         (only-in "compilation.rkt" compile CPS [ANF ANF*]))


;                       
;                       
;    ;;;;  ;;;;    ;; ; 
;   ;   ;   ;  ;  ;  ;; 
;   ;       ;  ;  ;     
;   ;       ;  ;   ;;;  
;   ;       ;;;       ; 
;   ;       ;         ; 
;   ;   ;   ;     ;;  ; 
;    ;;;   ;;;    ; ;;  
;                       
;                       
;                       
;                       

(define (eval e)
  (match (apply-reduction-relation -> e)
    [(list e*) (eval e*)]
    [(list) e]))

(define-syntax (test-CPS stx)
  (syntax-case stx ()
    [(_ expr expected)
     (with-syntax ([test-eq (syntax/loc stx
                              (test-equal actual (term expected)))]
                   [test-ty (syntax/loc stx
                              (test-predicate well-typed? actual))])
       (syntax 
        (let ([actual (term (CPS expr))])
          test-eq test-ty)))]))

(test-CPS
 (block
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
  (val x : (Int ε) = (shift f))
  (inc (dbl x)))
 (block
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
  (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
    (f (block (def g (x : (Int ε)) : (Int ε) = (k (block (inc (dbl x))))) g)))
  f1))

(test-CPS
 (reset
  (reset
   (block
    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
    (val x : (Int ε) = (shift f))
    (inc (dbl x)))))
 ((block
   (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
   (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
     (f (block (def g (x : (Int ε)) : (Int ε) = (k (block (inc (dbl x))))) g)))
   f1)
  (block (def id (x : (Int ε)) : (Int ε) = x) id)))

(test-CPS
 (shift
  (reset
   (block
    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ ((⇒ (Int ε) (Int ε)) ε) (Int ε)) ε) =
      (block
       (def g (k1 : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
         (k1 (k 7)))
       g))
    (val x : (Int ε) = (shift f))
    (inc (dbl x)))))
 ((block
   (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ ((⇒ (Int ε) (Int ε)) ε) (Int ε)) ε) =
     (block
      (def g (k1 : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
        (k1 (k 7)))
      g))
   (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ ((⇒ (Int ε) (Int ε)) ε) (Int ε)) ε) =
     (f (block (def g (x : (Int ε)) : (Int ε) = (k (block (inc (dbl x))))) g)))
   f1)
  (block (def id (x : (Int ε)) : (Int ε) = x) id)))

(let ([t (term (block
                (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
                (val x : (Int ε) = (shift f))
                (def g (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k x)))
                (val y : (Int ε) = (shift g))
                (inc (dbl y))))])
  (test-CPS 
   ,t
   (block
    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
    (def f1 (k1 : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
      (f (block (def g2 (x : (Int ε)) : (Int ε) = 
                  ((block
                    (def g (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k x)))
                    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
                      (g (block (def g1 (y : (Int ε)) : (Int ε) = (k (block (inc (dbl y))))) g1)))
                    f)
                   k1))
                g2)))
    f1))
  (test-equal (eval (compile (term (reset ,t)))) 63))

(test-CPS
 (block
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
  (val x : (Int ε) = (shift f))
  (val y : (Int ε) = (dbl x))
  (inc y))
 (block
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
  (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
    (f (block
        (def g (x : (Int ε)) : (Int ε) = 
          (k (block 
              (val y : (Int ε) = (dbl x))
              (inc y))))
        g)))
  f1))

(test-CPS
 (block
  (val w : (Int ε) = 3)
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k w)))
  (val x : (Int ε) = (shift f))
  (inc (dbl x)))
 (block
  (val w : (Int ε) = 3)
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k w)))
  (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
    (f (block (def g (x : (Int ε)) : (Int ε) = (k (block (inc (dbl x))))) g)))
  f1))

(test-CPS
 (block
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
  (val x : (Int ε) = (shift f))
  (block (inc (dbl x))))
 (block
  (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
  (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
    (f (block (def g (x : (Int ε)) : (Int ε) = (k (block (block (inc (dbl x)))))) g)))
  f1))

(test-CPS
 (block
  (def f (x : (Int ε)) : (Int (@cps Int Int)) =
    (block
     (def g (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
       (k (k x)))
     (shift g)))
  (val x : (Int ε) = (f 3))
  (inc (dbl x)))
 (block
  (def f (x : (Int ε)) : ((⇒ ((⇒ (Int ε) (Int ε)) ε) (Int ε)) ε) =
    (block
     (def g (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
       (k (k x)))
     g))
  (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
    ((f 3) 
     (block (def g (x : (Int ε)) : (Int ε) = (k (block (inc (dbl x))))) g)))
  f1))

(test-CPS
 ((reset
   (block
    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ (Int ε) (Int ε)) ε) = k)
    (val x : (Int ε) = (shift f))
    (inc (dbl x))))
  (reset
   (block
    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
    (val x : (Int ε) = (shift f))
    (inc (dbl x)))))
 (((block
    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ (Int ε) (Int ε)) ε) = k)
    (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ (Int ε) (Int ε)) ε) =
      (f (block (def g (x : (Int ε)) : (Int ε) = (k (block (inc (dbl x))))) g)))
    f1)
   (block (def id (x : (Int ε)) : (Int ε) = x) id))
  ((block
    (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 3)))
    (def f1 (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) =
      (f (block (def g (x : (Int ε)) : (Int ε) = (k (block (inc (dbl x))))) g)))
    f1)
   (block (def id (x : (Int ε)) : (Int ε) = x) id))))

(test-predicate
 well-typed?
 (term-let ([A (term ((⇒ (Int ε) (Int (@cps Int Int))) ε))]
            [B (term ((⇒ (Int ε) (Int (@cps Str Int))) ε))]
            [C (term ((⇒ (Int ε) (Int (@cps Int Str))) ε))]
            [i (term (Int ε))]
            [s (term (Str ε))])
           (term
            (CPS
             (block
              (def f1 (k : ((⇒ i i) ε)) : s = "q")
              (def g (x : i) : (Int (@cps Int Str)) =
                (shift f1))
              (def h (k : ((⇒ A B) ε)) : C = g)
              (def j (k : ((⇒ i s) ε)) : s = "q")
              (def k (x : i) : (Int (@cps Str Str)) =
                (shift j))
              (val x : A = (shift h))
              k)))))

(test-predicate
 well-typed?
 (term-let ([A (term ((⇒ (Int ε) (Int (@cps Int Int))) ε))]
            [C (term ((⇒ (Int ε) (Int (@cps Int Str))) ε))]
            [U (term ((⇒ (Int ε) (Int (@cps Str Int))) ε))]
            [V (term ((⇒ (Int ε) (Int (@cps Str Str))) ε))]
            [W (term ((⇒ (Int ε) (Str (@cps Int Int))) ε))]
            [i (term (Int ε))]
            [s (term (Str ε))])
           (term
            (CPS
             (block
              (def f1 (k : ((⇒ i i) ε)) : s = "q")
              (def f2 (x : i) : (Int (@cps Int Str)) =
                (shift f1))
              (def f3 (k : ((⇒ A W) ε)) : C = f2)
              (val x : A = (shift f3))
              (def f4 (k : ((⇒ s i) ε)) : i = 0)
              (def f5 (x : i) : (Str (@cps Int Int)) =
                (shift f4))
              (def f6 (k : ((⇒ U V) ε)) : W = f5)
              (shift f6))))))

(test-predicate
 well-typed?
 (term
  (block
   (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = 0)
   (def g (i : (Int ε)) : (Int (@cps Int Int)) = (shift f))
   (val x : ((⇒ (Int ε) (Int (@cps Int Int))) ε) = g)
   x)))

(test-predicate
 well-typed?
 (term
  (block
   (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = 0)
   (def g (i : (Int ε)) : (Int (@cps Int Int)) = (shift f))
   (def h (x : ((⇒ (Int ε) (Int (@cps Int Int))) ε)) 
     : ((⇒ (Int ε) (Int (@cps Int Int))) ε)
     = x)
   h)))

(test-equal (eval (compile shift-example-1)) 0)
(test-predicate well-typed? (compile shift-example-1))

(test-equal (eval (compile shift-example-2)) 1)
(test-predicate well-typed? (compile shift-example-2))

(test-equal (eval (compile shift-example-3)) 9)
(test-predicate well-typed? (compile shift-example-3))

(test-equal (eval (compile shift-example-4)) 2)
(test-predicate well-typed? (compile shift-example-4))

;                       
;                       
;    ;;   ;;; ;;; ;;;;;;
;     ;    ;;  ;   ;   ;
;    ; ;   ;;  ;   ; ;  
;    ; ;   ; ; ;   ;;;  
;    ; ;   ; ; ;   ; ;  
;   ;;;;;  ; ; ;   ;    
;   ;   ;  ;  ;;   ;    
;  ;;; ;;;;;; ;;  ;;;   
;                       
;                       
;                       
;                       

(define (ANF e)
  (unless (term (typeof () ,e))
    (error 'ANF "ill-typed ~s" e))
  (term (ANF* ,e)))

; app
(let ([t (term (block
                (def f (g : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ (Int ε) (Int ε)) ε) = g)
                ((f inc) (inc 7))))])
  (test-equal (ANF t) t))

(test-equal
 (ANF (term (block
             (def f (k : ((⇒ ((⇒ (Int ε) (Int ε)) ε) (Int ε)) ε)) : (Int ε) = 
               (dbl (k dbl)))
             (def g (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = 
               (dbl (k 7)))
             ((shift f) (shift g)))))
 (term (block
        (def f (k : ((⇒ ((⇒ (Int ε) (Int ε)) ε) (Int ε)) ε)) : (Int ε) = 
          (dbl (k dbl)))
        (def g (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = 
          (dbl (k 7)))
        (val x0 : ((⇒ (Int ε) (Int ε)) ε) = (shift f))
        (val x1 : (Int ε) = (shift g))
        (x0 x1))))

; block
(test-equal
 (ANF (term
       (block
        (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 0)))
        (f dbl)
        (inc (shift f))
        (block (shift f) 3)
        (shift f))))
 (term
  (block
   (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 0)))
   (f dbl)
   (val x0 : (Int ε) = (shift f))
   (inc x0)
   (val x2 : (Int ε) = 
        (block
         (val x1 : (Int ε) = (shift f))
         x1
         3))
   x2
   (shift f))))

; shift
(test-equal
 (ANF (term (block
             (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ ((⇒ (Str ε) (Str ε)) ε) (Str ε)) ε) = 
               (block (def g (h : ((⇒ (Str ε) (Str ε)) ε)) : (Str ε) = (h "s")) g))
             (shift (reset (dbl (shift f)))))))
 (term (block
        (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : ((⇒ ((⇒ (Str ε) (Str ε)) ε) (Str ε)) ε) = 
          (block (def g (h : ((⇒ (Str ε) (Str ε)) ε)) : (Str ε) = (h "s")) g))
        (shift (reset (block (val x0 : (Int ε) = (shift f)) (dbl x0)))))))

; reset
(test-equal
 (ANF (term (block
             (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 7)))
             (reset (block (shift f) (shift f))))))
 (term (block
        (def f (k : ((⇒ (Int ε) (Int ε)) ε)) : (Int ε) = (k (k 7)))
        (reset 
         (block
          (val x0 : (Int ε) = (shift f))
          x0
          (shift f))))))

; primitives, strings, and numbers
(for-each (λ (t) (test-equal (ANF t) t))
          '(inc "s" 7.5))

(test-results)