#lang racket

(require redex "grammar.rkt")

(provide (all-defined-out))

(define run-gadt
  (reduction-relation
   gadt-lang
   
   [==> ((λ x e_2) e_1)
        (subst {[x := e_1]} e_2)]
   
   [==> (case e_0 of
          [(K_1 x_1 ...) → e_1] ...
          [(K_2 x_2 ...) → e_2]
          [(K_3 x_3 ...) → e_3] ...)
        (subst {[x_2 := e] ...} e_2)
        (where (K e ...) (flatten-application e_0))
        (side-condition (term (≡ K K_2)))
        (side-condition (term (≈ (x_2 ...) (e ...))))]
   
   [==> (let (x = e_1) in e_2)
        (subst
         {[x := (subst
                 {[x := (let (x = e_1) in x)]}
                 e_1)]}
         e_2)]
   
   [==> (let (x :: σ = e_1) in e_2)
        (subst
         {[x := (subst
                 {[x := (let (x :: σ = e_1) in x)]}
                 e_1)]}
         e_2)]
   
   with
   
   [(--> (program dd ... (in-hole κ exp1)) (program dd ... (in-hole κ exp2)))
    (==> exp1 exp2)]
   
   [(--> (in-hole κ exp1) (in-hole κ exp2))
    (==> exp1 exp2)]
   
   ))

(define (test-eval)
  (redex-test-suite
    (test-->> run-gadt (term ((λ x x) Make)) (term Make))
    (test-->> run-gadt (term (case Make of [(Make) → MT])) (term MT))
    (test-->> run-gadt (term (let (x = Make) in x)) (term Make))
    (test-->> run-gadt
      (term (let (x :: (∀ () [^] ⇒ (Unit)) = Make) in x))
      (term Make))))
